#!/usr/bin/env python

# standard library
import time

# external libraries
from numpy import hstack, ones, pi, linspace, array, zeros, zeros_like
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from pydy.models import n_link_pendulum_on_cart
from sympy import symbols


def run_benchmark(max_num_links, num_time_steps=1000):
    """Runs the n link pendulum derivation, code generation, and integration
    for each n up to the max number provided and generates a plot of the
    results."""

    methods = ['lambdify', 'theano', 'cython']

    link_numbers = range(1, max_num_links + 1)

    derivation_times = zeros(len(link_numbers))
    integration_times = zeros((max_num_links, len(methods)))
    code_generation_times = zeros_like(integration_times)

    for j, n in enumerate(link_numbers):

        title = "Pendulum with {} links.".format(n)
        print(title)
        print('=' * len(title))

        start = time.time()
        sys = n_link_pendulum_on_cart(n, cart_force=False)

        m = symbols('m:{}'.format(n + 1))
        l = symbols('l:{}'.format(n))
        g = symbols('g')

        derivation_times[j] = time.time() - start
        print('The derivation took {:1.5f} seconds.\n'.format(derivation_times[j]))

        # Define the numerical values: parameters, time, and initial conditions
        arm_length = 1. / n
        bob_mass = 0.01 / n
        parameter_vals = [9.81, 0.01 / n]
        for i in range(n):
            parameter_vals += [arm_length, bob_mass]

        times = linspace(0, 10, num_time_steps)
        sys.times = times

        x0 = hstack(
            (0,
             pi / 2 * ones(len(sys.coordinates) - 1),
             1e-3 * ones(len(sys.speeds))))
        sys.initial_conditions = dict(zip(sys.states, x0))

        constants = [g, m[0]]
        for i in range(n):
            constants += [l[i], m[i + 1]]

        sys.constants = dict(zip(constants, array(parameter_vals)))

        for k, method in enumerate(methods):

            subtitle = "Generating with {} method.".format(method)
            print(subtitle)
            print('-' * len(subtitle))
            start = time.time()
            sys.generate_ode_function(generator=method)
            code_generation_times[j, k] = time.time() - start
            print('The code generation took {:1.5f} seconds.'.format(
                code_generation_times[j, k]))

            start = time.time()
            sys.integrate()
            integration_times[j, k] = time.time() - start
            print('ODE integration took {:1.5f} seconds.\n'.format(
                integration_times[j, k]))

        del sys

    # plot the results
    fig, ax = plt.subplots(3, 1, sharex=True)

    ax[0].plot(link_numbers, derivation_times)
    ax[0].set_title('Symbolic Derivation Time')

    ax[1].plot(link_numbers, code_generation_times)
    ax[1].set_title('Code Generation Time')
    ax[1].legend(methods, loc=2)

    ax[2].plot(link_numbers, integration_times)
    ax[2].set_title('Integration Time')
    ax[2].legend(methods, loc=2)

    for a in ax.flatten():
        a.set_ylabel('Time [s]')

    ax[-1].set_xlabel('Number of links')

    plt.tight_layout()

    fig.savefig('benchmark-results.png')

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(
        description='Run the n link pendulum benchmark.')

    parser.add_argument('max_num_links', type=int,
        help="The maximum number of links to compute.")

    parser.add_argument('num_time_steps', type=int,
        help="The number of integration time steps.")

    args = parser.parse_args()

    run_benchmark(args.max_num_links, args.num_time_steps)



"""
This is a script to test the time taken by the generated
rhs function in ode_function_generators.

We allow the user to specify what type, the extra args
should be for both the constants and the specifieds. The constants
can be None, 'array', or 'dictionary'. The specifieds can be None,
'array', 'function', or 'dictionary'. Thus, we have 12 permutations.

The script outputs the average time taken to numerically
evaluate the right hand side of the first order differential
equation for each of the 12 permutations.

"""

import timeit

import numpy as np
import sympy as sm

from pydy import models
from pydy.codegen.ode_function_generators import LambdifyODEFunctionGenerator

sys = models.n_link_pendulum_on_cart(3, True, True)

right_hand_side = sys.eom_method.rhs()

constants = list(sm.ordered(sys.constants_symbols))

specifieds = list(sm.ordered(sys.specifieds_symbols))

constants_arg_types = ['array', 'dictionary']
specifieds_arg_types = ['array', 'function', 'dictionary']

p_array = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
p_dct = dict(zip(constants, p_array))

p = {}

p['array'] = p_array
p['dictionary'] = p_dct

r_array = np.array([1.0, 2.0, 3.0, 4.0])
# r_dct_1 = dict(zip(specifieds, r_array))
# r_dct_2 = {tuple(specifieds):
#           lambda x, t: r_array}
r_dct_3 = {specifieds[0]: lambda x, t: np.ones(1),
           (specifieds[3], specifieds[1]):
           lambda x, t: np.array([4.0, 2.0]),
           specifieds[2]: 3.0 * np.ones(1)}
r_func = lambda x, t: np.array([1.0, 2.0, 3.0, 4.0])

r = {}

r['array'] = r_array
r['dictionary'] = r_dct_3
r['function'] = r_func

x = np.array([0.125, 0.250, 0.375, 0.456,
              0.123, 0.457, 0.999, 0.192])
itr = 1000
print('The time taken by rhs function in {} iterations'.format(itr))

for p_arg_type in constants_arg_types:
    for r_arg_type in specifieds_arg_types:

        g = LambdifyODEFunctionGenerator(right_hand_side,
                                         sys.coordinates,
                                         sys.speeds,
                                         constants,
                                         specifieds=specifieds,
                                         constants_arg_type=p_arg_type,
                                         specifieds_arg_type=r_arg_type)

        rhs = g.generate()

        time = timeit.repeat("rhs(x, 0.0, r[r_arg_type], p[p_arg_type])",
                             "from __main__ import rhs,x,r,p,"
                             "r_arg_type,p_arg_type",
                             number=itr)

        print('For constants argument type - "{p_arg}" and '
              'specifieds argument type - "{r_arg}" is '
              .format(p_arg=p_arg_type, r_arg=r_arg_type))

        print(sum(time)/3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages

exec(open('pydy/version.py').read())

# I was getting the same error as:
# https://github.com/statsmodels/statsmodels/issues/1073, so the following
# line is added.
os.environ["MPLCONFIGDIR"] = "."


install_requires=['numpy>=1.11.0',
                  'scipy>=0.17.1',
                  'sympy>=0.7.6.1']

if os.name == 'nt':
    install_requires.append('PyWin32>=219')

setup(
    name='pydy',
    version=__version__,
    author='PyDy Authors',
    author_email='pydy@googlegroups.com',
    url="http://pydy.org",
    description='Python tool kit for multi-body dynamics.',
    long_description=open('README.rst').read(),
    keywords="multibody dynamics",
    license='LICENSE.txt',
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={'doc': ['sphinx', 'numpydoc'],
                    'codegen': ['Cython>=0.23.4',
                                'Theano>=0.8.0'],
                    'examples': ['matplotlib>=1.5.1',
                                 'notebook>=4.0.0,<5.0.0',
                                 'ipywidgets>=4.0.0,<5.0.0'],
                    },
    tests_require=['nose>=1.3.7'],
    test_suite='nose.collector',
    include_package_data=True,
    classifiers=['Development Status :: 4 - Beta',
                 'Intended Audience :: Science/Research',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Scientific/Engineering :: Physics',
                 ],
)
